------------------------------------------------------------------------
-- agentDoorBreak
--
-- Contains code for a custom version of breakDoor that is intended to be
-- used by agents. Use by guards probably will work, but behavior in this case is not defined.

local abilitydefs = include( "sim/abilitydefs" )
local abilityutil = include("sim/abilities/abilityutil")
local breakDoor = abilitydefs.lookupAbility("breakDoor")
local cdefs = include("client_defs")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")
local speechdefs = include("sim/speechdefs")
local util = include("client_util")

local function calculateAgentDoorBreakKOandArmorPiercing(agentUnit)
	-- Calculate doorBreakKOTime and armorPiercing
	local doorBreakKOTime = 0
	local armorPiercing = 0

	-- Add bonuses specific to door breaking:
	doorBreakKOTime = doorBreakKOTime + (agentUnit:getTraits().doorBreakKOBonus or 0)
	armorPiercing = armorPiercing + (agentUnit:getTraits().doorBreakArmorPiercingBonus or 0)

	-- Add melee bonuses
	doorBreakKOTime = doorBreakKOTime + (agentUnit:getTraits().meleeDamage or 0)
	armorPiercing = armorPiercing + agentUnit:countArmorPiercingUpgrades("addArmorPiercingMelee")

	-- Kinetic Capacitor bonus
	if agentUnit:getTraits().kinetic_capacitor_bonus and agentUnit:getTraits().kinetic_capacitor_bonus_charged  then
		doorBreakKOTime = doorBreakKOTime + agentUnit:getTraits().kinetic_capacitor_bonus
		armorPiercing = armorPiercing + agentUnit:getTraits().kinetic_capacitor_bonus
	end

	-- Sharp's augment bonus
	if agentUnit:countAugments("augment_sharp_1") > 0 then
		local augmentCount = agentUnit:getAugmentCount()
		doorBreakKOTime = doorBreakKOTime + math.floor(augmentCount / 3)
	end
	if agentUnit:countAugments("augment_sharp_2") > 0 then
		local augmentCount = agentUnit:getAugmentCount()
		armorPiercing = armorPiercing + math.floor(augmentCount / 3)
	end

	return doorBreakKOTime, armorPiercing
end

local function getImpactCells(firstCell, dir, agentUnit)
	local curCell = firstCell
	local fullImpactCells = {}
	local halfImpactCells = {}
	for i=1,(agentUnit:getTraits().doorBreakFullImpactRange or 0) do
		if curCell then
			fullImpactCells[i] = curCell
			if curCell.exits[dir] then
				curCell = curCell.exits[dir].cell
			else
				curCell = nil
			end
		else
			break
		end
	end
	for i=1,(agentUnit:getTraits().doorBreakHalfImpactRange or 0) do
		if curCell then
			halfImpactCells[i] = curCell
			if curCell.exits[dir] then
				curCell = curCell.exits[dir].cell
			else
				curCell = nil
			end
		else
			break
		end
	end
	return fullImpactCells, halfImpactCells
end

-- Handles breaking the door and K.O.ing anything on the other side
-- Note that most of this code is copy pasted from engine.lua modifyExit()
local function customModifyExit(sim, cell, dir, agentUnit)
	local exit = cell.exits[dir]
	assert( exit )

	local rdir = simquery.getReverseDirection( dir )
	local reverseExit = exit.cell.exits[ rdir ]
	assert( reverseExit )

	if agentUnit then
		agentUnit:getTraits().modifyingExit = exit
	end

	-- Get cells that will be impacted by doorBreak
	local fullImpactCells, halfImpactCells = getImpactCells(cell.exits[dir].cell, dir, agentUnit)

	-- Break the door open, permanently
	assert(exit.door and reverseExit.door)
	assert(exit.closed and reverseExit.closed)

	exit.closed, reverseExit.closed = nil, nil
	exit.locked, reverseExit.locked = true, true
	sim:getLOS():removeSegments( cell, dir, exit.cell, rdir )
	sim:emitSound(exit.breakSound, cell.x, cell.y, agentUnit, {{x=fullImpactCells[1].x,y=fullImpactCells[1].y}})
	if sim:getCurrentPlayer() ~= sim:getNPC() then
		sim:getNPC().pather:invalidatePathsCrossingCell(cell.x, cell.y)
		for _, fullCell in ipairs(fullImpactCells) do
			sim:getNPC().pather:invalidatePathsCrossingCell(fullCell.x, fullCell.y)
		end
		for _, halfCell in ipairs(halfImpactCells) do
			sim:getNPC().pather:invalidatePathsCrossingCell(halfCell.x, halfCell.y)
		end
	end

	sim:dispatchEvent(simdefs.EV_EXIT_MODIFIED, { cell = cell, dir = dir, exitOp = simdefs.EXITOP_BREAK_DOOR })

	-- Refresh LOS for any unit whose LOS arc intersects the exit.
	--use the door before refreshing LOS
	sim:triggerEvent(simdefs.TRG_UNIT_USEDOOR_PRE, { exitOp = simdefs.EXITOP_BREAK_DOOR, cell = cell, tocell = exit.cell, unit = agentUnit })

			local seers = {}
			for unitID, seerUnit in pairs(sim:getAllUnits()) do
				local x0, y0 = seerUnit:getLocation()
					if seerUnit:getTraits().hasSight and x0 then
							local nx, ny = exit.cell.y - cell.y, exit.cell.x - cell.x  -- normal of exit.cell -> exit.cell
							local cx, cy = (cell.x + exit.cell.x) / 2, (cell.y + exit.cell.y) / 2
							if (x0 == cell.x and y0 == cell.y) or (x0 == exit.cell.x and y0 == exit.cell.y) or
								sim:getLOS():withinLOS( seerUnit, cx + nx/2, cy + ny/2, cx - nx/2, cy - ny/2 ) then
									table.insert( seers, unitID )
							end
					end
			end

	for _, seerID in ipairs(seers) do
		local seerUnit
		if seerID >= simdefs.SEERID_PERIPHERAL then
			seerUnit = sim:getUnit(seerID - simdefs.SEERID_PERIPHERAL)
		else
			seerUnit = sim:getUnit(seerID)
		end

		if seerUnit then
			if seerUnit:hasTrait("hasSight") then
				sim:refreshUnitLOS(seerUnit)
			end

			-- Make sure to also refresh the exit knowledge on either side of the exit.
			if seerUnit:getPlayerOwner() then
				seerUnit:getPlayerOwner():glimpseExit(cell.x, cell.y, dir)
			end
		end
	end

	sim:triggerEvent(simdefs.TRG_UNIT_USEDOOR, { exitOp = simdefs.EXITOP_BREAK_DOOR, cell = cell, tocell = exit.cell, unit = agentUnit })

	-- K.O. folks on the other side of the door based on STRENGTH and melee augments
	local doorBreakKOTime, armorPiercing = calculateAgentDoorBreakKOandArmorPiercing(agentUnit)
	if doorBreakKOTime > 0 then
		-- Do full impact cells
		for _, fullCell in ipairs(fullImpactCells) do
			for _,targetUnit in ipairs(fullCell.units) do
				if simquery.isAgent(targetUnit) and not targetUnit:isDead() and not targetUnit:getTraits().isDrone == true then

					-- Determine targets armor and K.O. damage after K.O. resistance
					local targetArmor = targetUnit:getArmor()
					local koTime = targetUnit:processKOresist(doorBreakKOTime)

					-- Check if we can actually K.O. the target
					local targetKOTime = 0
					if targetUnit:isKO() then
						targetKOTime = targetUnit:getKOTimer()
					end
					if armorPiercing >= targetArmor and koTime > targetKOTime then
						-- K.O. the target
						sim:dispatchEvent(simdefs.EV_UNIT_FLOAT_TXT, {txt=STRINGS.UI.FLY_TXT.DOOR_SLAMMED,x=fullCell.x,y=fullCell.y,color={r=163/255,g=0,b=0,a=1},alwaysShow=true})
						if targetUnit:getTraits().canKO then
							targetUnit:setKO(sim, koTime)
						elseif targetUnit:getTraits().isDrone and targetUnit.deactivate then
							targetUnit:deactivate(sim)
						end
					end
				end
			end
		end
		-- Do half impact cells
		doorBreakKOTime = math.ceil(doorBreakKOTime / 2)
		armorPiercing = math.ceil(armorPiercing / 2)
		for _, halfCell in ipairs(halfImpactCells) do
			for _,targetUnit in ipairs(halfCell.units) do
				if simquery.isAgent(targetUnit) and not targetUnit:isDead() and not targetUnit:getTraits().isDrone == true then

					-- Determine targets armor and K.O. damage after K.O. resistance
					local targetArmor = targetUnit:getArmor()
					local koTime = targetUnit:processKOresist(doorBreakKOTime)

					-- Check if we can actually K.O. the target
					local targetKOTime = 0
					if targetUnit:isKO() then
						targetKOTime = targetUnit:getKOTimer()
					end
					if armorPiercing >= targetArmor and koTime > targetKOTime then
						-- K.O. the target
						sim:dispatchEvent(simdefs.EV_UNIT_FLOAT_TXT, {txt=STRINGS.UI.FLY_TXT.DOOR_SLAMMED,x=halfCell.x,y=halfCell.y,color={r=163/255,g=0,b=0,a=1},alwaysShow=true})
						if targetUnit:getTraits().canKO then
							targetUnit:setKO(sim, koTime)
						elseif targetUnit:getTraits().isDrone and targetUnit.deactivate then
							targetUnit:deactivate(sim)
						end
					end
				end
			end
		end
	end

	if agentUnit and agentUnit:isValid() and not agentUnit:getTraits().interrupted then
		sim:processReactions(agentUnit)
	end

	if agentUnit then
		agentUnit:getTraits().modifyingExit = nil
	end
end

-- Copied from hud/combat_panel.lua
local COLOR_BLUE_DARK = util.color(50/255,87/255,87/255,150/255)
local COLOR_BLUE_LIGHT = util.color(150/255,255/255,255/255,1)

local COLOR_RED_DARK = util.color(80/255,0/255,0/255,150/255)
local COLOR_RED_LIGHT = util.color(163/255,0/255,0/255,1)

local function createCombatPanel(hud, koDamage, armorPiercing, targetUnit)
	local wx, wy = hud._game:cellToWorld(targetUnit:getLocation())
	local combatPanelWidget = hud._world_hud:createWidget(hud._world_hud.HUD_FLOATERS, "eaa_combatPanel", {worldx = wx, worldy = wy, worldz = 0})--, updateFn, destroyFn )
	combatPanelWidget._screen = hud._world_hud._screen

	local color, textColor = COLOR_BLUE_DARK, COLOR_BLUE_LIGHT
	if armorPiercing >= targetUnit:getArmor() then
		koDamage = targetUnit:processKOresist(koDamage)
		combatPanelWidget.binder.damageTxt:setText(util.sformat(STRINGS.FORMATS.AMT_KO, koDamage))
		combatPanelWidget.binder.targetTxt:setText(targetUnit:getName())
	else
		color, textColor = COLOR_RED_DARK, COLOR_RED_LIGHT
		combatPanelWidget.binder.damageTxt:setText(STRINGS.UI.COMBAT_PANEL_ARMORED)
		combatPanelWidget.binder.targetTxt:setText(targetUnit:getName())
	end

	combatPanelWidget:setVisible(true)
	combatPanelWidget.binder.hitBackground:setColor(color:unpack())
	combatPanelWidget.binder.targetTxt:setColor(textColor:unpack())
	combatPanelWidget.binder.damageTxt:setColor(textColor:unpack())

	combatPanelWidget.eaa_old_setPosition = combatPanelWidget.setPosition
	combatPanelWidget.setPosition = function(self, x0, y0)
		local width,_ = self._screen:getResolution()
		local offset = 100/width
		if x0 > 0.70 then
			offset=  - 100/width
		end
		y0 = math.min(math.max(y0, 0.15), 0.7)
		-- self:eaa_old_setPosition(x0-0.5+offset, y0-0.15)
		self:eaa_old_setPosition(x0-0.53+offset,y0-0.11)
	end

	return combatPanelWidget
end

local agentDoorBreak_tooltip = class(util.tooltip)

-- ability -> the ability itself (agentDoorBreak)
-- hud
-- sim
-- agentUnit -> the agent that has the ability
-- tooltipBody -> the message to display in the tooltip
-- x -> the x coordinate of the cell the agent is standing on
-- y -> the y coordinate of the cell the agent is standing on
-- dir -> the direction that the exit is in
function agentDoorBreak_tooltip:init(ability, hud, sim, agentUnit, x, y, dir)
	-- Initialize self
	util.tooltip.init(self, hud._screen, hud._game)
	local section = util.tooltip.addSection(self)
	self._agentUnit = agentUnit
	self._fullImpactTargets = {}
	self._halfImpactTargets = {}
	self._combatPanels = {}

	-- Add name and description
	section:addLine("<ttheader>"..STRINGS.ABILITIES.BREAK_DOOR.."</>\n"..STRINGS.ABILITIES.BREAK_DOOR_DESC)
	-- Add range, K.O., and armor piercing information
	local fullImpactRange = agentUnit:getTraits().doorBreakFullImpactRange or 0
	local halfImpactRange = agentUnit:getTraits().doorBreakHalfImpactRange or 0
	local koDamage, armorPiercing = calculateAgentDoorBreakKOandArmorPiercing(agentUnit)

	local txt_fullImpactRange = fullImpactRange == 1 and "Tile" or tostring(fullImpactRange).." Tiles"
	local txt_halfImpactRange = halfImpactRange == 1 and "Tile" or tostring(halfImpactRange).." Tiles"

	section:addLine("<c:FF8411>First "..txt_fullImpactRange.."</c>")
	section:addAbility(STRINGS.ITEMS.TOOLTIPS.KO_DAMAGE, util.sformat(STRINGS.ITEMS.TOOLTIPS.KO_DAMAGE_DESC, koDamage), "gui/icons/arrow_small.png")
	if armorPiercing > 0 then
		section:addAbility(STRINGS.ITEMS.TOOLTIPS.ARMOR_PIERCING, string.format(STRINGS.ITEMS.TOOLTIPS.ARMOR_PIERCING_DESC, armorPiercing), "gui/icons/arrow_small.png")
	end

	if halfImpactRange > 0 then
		koDamage = math.ceil(koDamage/2)
		armorPiercing = math.ceil(armorPiercing/2)
		section:addLine("<c:FF8411>Next "..txt_halfImpactRange.."</>")
		section:addAbility(STRINGS.ITEMS.TOOLTIPS.KO_DAMAGE, util.sformat(STRINGS.ITEMS.TOOLTIPS.KO_DAMAGE_DESC, koDamage), "gui/icons/arrow_small.png")
		if armorPiercing > 0 then
			section:addAbility(STRINGS.ITEMS.TOOLTIPS.ARMOR_PIERCING, string.format(STRINGS.ITEMS.TOOLTIPS.ARMOR_PIERCING_DESC, armorPiercing), "gui/icons/arrow_small.png")
		end
	end

	-- Add Warnings
	if simquery.isSecurityExit(sim:getCell(x,y).exits[dir]) then
		section:addRequirement("Raises alarm by 1")
	end

	-- Add reasons we can't use the ability (if there are any)
	local enabled, reason = ability:canUseAbility(sim, agentUnit)
	if reason then
		section:addRequirement(reason)
	end

	-- Find valid K.O. targets
	if enabled == true then
		local fullImpactCells, halfImpactCells = getImpactCells(sim:getCell(x,y).exits[dir].cell, dir, agentUnit)
		-- Do full impact cells
		for _, cell in ipairs(fullImpactCells) do
			for _,targetUnit in ipairs(cell.units) do
				if simquery.isAgent(targetUnit) and not targetUnit:getTraits().isDrone == true and not targetUnit:isDead() and sim:canPlayerSeeUnit(sim:getPC(), targetUnit) then
					table.insert(self._fullImpactTargets, targetUnit)
				end
			end
		end
		-- Do half impact cells
		for _, cell in ipairs(halfImpactCells) do
			for _,targetUnit in ipairs(cell.units) do
				if simquery.isAgent(targetUnit) and not targetUnit:getTraits().isDrone == true and not targetUnit:isDead() and sim:canPlayerSeeUnit(sim:getPC(), targetUnit) then
					table.insert(self._halfImpactTargets, targetUnit)
				end
			end
		end
	end
end

function agentDoorBreak_tooltip:activate(screen)
	util.tooltip.activate(self, screen)
	self._game.hud:previewAbilityAP(self._agentUnit, simdefs.EAA_DOOR_BREAK_MP_COST)
	-- Highlight targets that will get K.O.
	local koDamage, armorPiercing = calculateAgentDoorBreakKOandArmorPiercing(self._agentUnit)
	for _, targetUnit in ipairs(self._fullImpactTargets) do
		table.insert(self._combatPanels, createCombatPanel(self._game.hud, koDamage, armorPiercing, targetUnit))
		self._game.boardRig:getUnitRig(targetUnit:getID())._prop:setRenderFilter(cdefs.RENDER_FILTERS["focus_highlite"])
	end
	koDamage = math.ceil(koDamage/2)
	armorPiercing = math.ceil(armorPiercing/2)
	for _, targetUnit in ipairs(self._halfImpactTargets) do
		table.insert(self._combatPanels, createCombatPanel(self._game.hud, koDamage, armorPiercing, targetUnit))
		self._game.boardRig:getUnitRig(targetUnit:getID())._prop:setRenderFilter(cdefs.RENDER_FILTERS["focus_highlite"])
	end
end

function agentDoorBreak_tooltip:deactivate()
	util.tooltip.deactivate(self)
	self._game.hud:previewAbilityAP(self._agentUnit, 0)
	for _, targetUnit in ipairs(self._fullImpactTargets) do
		self._game.boardRig:getUnitRig(targetUnit:getID())._prop:setRenderFilter(nil)
	end
	for _, targetUnit in ipairs(self._halfImpactTargets) do
		self._game.boardRig:getUnitRig(targetUnit:getID())._prop:setRenderFilter(nil)
	end
	for _, combatPanelWidget in ipairs(self._combatPanels) do
		self._game.hud._world_hud:destroyWidget(self._game.hud._world_hud.HUD_FLOATERS, combatPanelWidget)
	end
	self._combatPanels = {}
end

local agentDoorBreak = util.extend(breakDoor)
{
	-- getName() is required when you use an exitTarget
	getName = function()--( self, sim, unit)
		return breakDoor.name
	end,
	HUDpriority = 1001,

	createToolTip = nil, -- Remove the createTooltip function that is included with breakDoor since we are using onTooltip instead
	--ability:onTooltip( self._hud, sim, abilityOwner, abilityUser, unpack(abilityTargets)
	onTooltip = function(self, hud, sim, abilityOwner, abilityUser, x, y, dir)
		return agentDoorBreak_tooltip(self, hud, sim, abilityUser, x, y, dir)
	end,

	-- Finds only eligible targets for the ability. In this case, any open exit
	-- Note that canUseAbility() is called by the game BEFORE and after acquireTargets()
	-- If canUseAbility returns false, then acquireTargets() is not called.
	-- showTargets = function( self, targets, game, sim, abilityOwner, abilityUser)
	acquireTargets = function(self, targets, game, sim, abilityOwner, abilityUser)
		local exits = {}

		if abilityOwner == abilityUser then
			local fromCell = sim:getCell(abilityUser:getLocation())
			for dir, exit in pairs(fromCell.exits) do
				if simquery.isClosedDoor(exit) and
				   (not simquery.isSecurityExit(exit) or (abilityOwner:countAugments("eaa_augment_steel_foot") > 0)) then
					table.insert(exits, {x = fromCell.x, y = fromCell.y, dir = dir})
				end
			end
		end

		return targets.exitTarget(game, exits, self, abilityOwner, abilityUser)
	end,

	canUseAbility = function(self, sim, unit, userUnit, x, y, direction)
		if not simquery.isAgent(unit) then
			return false
		end

		if unit:getPlayerOwner() and unit:getPlayerOwner():isPC() and unit:getMP() < simdefs.EAA_DOOR_BREAK_MP_COST then
			return false, string.format(STRINGS.UI.REASON.REQUIRES_AP, simdefs.EAA_DOOR_BREAK_MP_COST)
		end

		local fromCell = sim:getCell(x, y) or sim:getCell( unit:getLocation() )
		local doorExit = direction and fromCell.exits[direction] or nil

		if not doorExit then
			for dir, exit in pairs(fromCell.exits) do
				if simquery.isClosedDoor(exit) and
				   (not simquery.isSecurityExit(exit) or (unit:countAugments("eaa_augment_steel_foot") > 0)) then
					doorExit = exit
					direction = dir
				end
			end
		end

		if not doorExit then
			return false
		end

		local canModifyExit, reason = simquery.canModifyExit( unit, simdefs.EXITOP_BREAK_DOOR, fromCell, direction)
		if not canModifyExit and reason ~= STRINGS.UI.DOORS.SECURITY then
			return false
		end

		return abilityutil.checkRequirements( unit, userUnit )
	end,

	executeAbility = function(self, sim, unit, userUnit, x, y, direction)
		-- Only agents get charged MP, do not charge guards.
		if userUnit:getBrain() == nil then
			userUnit:useMP(simdefs.EAA_DOOR_BREAK_MP_COST, sim)
		end

		unit:resetAllAiming()

		local cell = sim:getCell(x, y) or sim:getCell( unit:getLocation() )
		assert( cell )
		local doorExit = direction and cell.exits[direction] or nil

		if not doorExit then
			for dir, exit in pairs(cell.exits) do
				if simquery.isClosedDoor(exit) and
				   (not simquery.isSecurityExit(exit) or (userUnit:countAugments("eaa_augment_steel_foot") > 0)) then
					doorExit, direction = exit, dir
				end
			end
		end
		assert(doorExit)
		local canModifyExit, reason = simquery.canModifyExit(unit, simdefs.EXITOP_BREAK_DOOR, cell, direction)
		assert(canModifyExit or reason == STRINGS.UI.DOORS.SECURITY)
		assert(simquery.canReachDoor(unit, cell, direction))


		simquery.suggestAgentFacing(userUnit, direction)
		if userUnit:isNPC() then
			sim:emitSpeech(unit, speechdefs.HUNT_BREAKDOOR)
		end
		if not unit:getTraits().noDoorAnim then
			sim:dispatchEvent(simdefs.EV_UNIT_USEDOOR, { unitID = unit:getID(), facing = direction, exitOp=simdefs.EXITOP_BREAK_DOOR })
		end
		sim:getPC():glimpseUnit(sim, unit:getID())
		sim:startTrackerQueue(true)
		sim:startDaemonQueue()
		sim:dispatchEvent(simdefs.EV_KO_GROUP, true)
		if simquery.isSecurityExit(doorExit) then
			sim:dispatchEvent(simdefs.EV_UNIT_FLOAT_TXT, {txt=util.sformat(STRINGS.UI.ALARM_ADD, 1), x=doorExit.cell.x, y=doorExit.cell.y, color = {r=255/255, g=10/255, b=10/255,a=1}})
			sim:trackerAdvance(1)
		end
		customModifyExit(sim, cell, direction, unit)
		--sim:modifyExit(cell, direction, simdefs.EXITOP_BREAK_DOOR, unit, false)
		if unit:isValid() and not unit:getTraits().noDoorAnim then
			sim:dispatchEvent(simdefs.EV_UNIT_USEDOOR_PST, { unitID = unit:getID(), facing = direction, exitOp=simdefs.EXITOP_BREAK_DOOR })
		end
		sim:dispatchEvent(simdefs.EV_KO_GROUP, false )
		sim:startTrackerQueue(false)
		sim:processDaemonQueue()
		if unit:isValid() and not unit:getTraits().noDoorAnim and not unit:getTraits().interrupted then
			sim:dispatchEvent(simdefs.EV_UNIT_GUNCHECK, { unit = unit, facing = direction })
		end

	end
}

return agentDoorBreak
