----------------------------------------------------------------------
-- The door slam ability. Manages slamming doors and making the noise that alerts guards
--

-- Will need to add this ability to the list of default agent abilities, see sim\unitdefs\agentdefs.lua and sim\unitdefs\commondefs.lua

local abilityutil = include("sim/abilities/abilityutil")
local simactions = include("sim/simactions")
local simdefs = include("sim/simdefs")
local simquery = include("sim/simquery")

local doorSlam_tooltip = class(abilityutil.hotkey_tooltip)

function doorSlam_tooltip:init(hud, unit, ...)
	abilityutil.hotkey_tooltip.init(self, ...)
	self._game = hud._game
	self._unit = unit
end

function doorSlam_tooltip:activate(screen)
	abilityutil.hotkey_tooltip.activate(self, screen)
	self._game.hud:previewAbilityAP(self._unit, simdefs.EAA_DOOR_SLAM_MP_COST)
end

function doorSlam_tooltip:deactivate()
	abilityutil.hotkey_tooltip.deactivate(self)
	self._game.hud:previewAbilityAP(self._unit, 0)
end

local doorSlam =
{
	name = STRINGS.EXTRA_AGENT_ABILITIES.TOOLTIPS.DOOR_SLAM,
	profile_icon = "gui/icons/icon-action_door_slam_small.png",
	HUDpriority = 1001,
	proxy = true, -- Tells the program to show on target (instead of the agent/owner)
	alwaysShow = nil, -- When true, shows a grayed out version of the ability when it can't be used
	ghostable = nil, -- Ghosts are red images that show you where something you saw was last.

	-- Allows us to customize the name if necessary
	getName = function(self, sim, unit)
		return self.name
	end,

	-- Creates the tooltip. May need to tweak this if the door is open or destroyed
	onTooltip = function(self, hud, sim, abilityOwner, abilityUser)
		return doorSlam_tooltip(hud, abilityUser, self, sim, abilityOwner, STRINGS.EXTRA_AGENT_ABILITIES.TOOLTIPS.DOOR_SLAM_DESC)
	end,
	--createToolTip = function()--(self, sim, unit, targetCell)
	--	return abilityutil.formatToolTip(STRINGS.EXTRA_AGENT_ABILITIES.TOOLTIPS.DOOR_SLAM, STRINGS.EXTRA_AGENT_ABILITIES.TOOLTIPS.DOOR_SLAM_DESC )
	--end,

	-- Finds only eligible targets for the ability. In this case, any open exit
	acquireTargets = function(self, targets, game, sim, abilityOwner, abilityUser)
		-- Code to find exits copied from the base games scripts\sim\abilities\peek.lua file
		-- Most comments added by myself as I tried to understand what it was doing
		local exits = {} -- This is a table where we will store our exits (doors)

		-- We need to make sure that abilityOwner and abilityUser are the same agent, otherwise we might display another agents "doorSlam" ability (which would result in 2 "doorSlam" icons when two agents are adjacent to each other)
		if abilityOwner == abilityUser then
			local fromCell = sim:getCell(abilityUser:getLocation()) -- Get the cell that the agent is standing on

			-- Add all open doors within reach of the agent to the list of exits
			for i = 1, #simdefs.ADJACENT_EXITS, 3 do
				local dx, dy, dir = simdefs.ADJACENT_EXITS[i], simdefs.ADJACENT_EXITS[i+1], simdefs.ADJACENT_EXITS[i+2]
				local cell = sim:getCell(fromCell.x + dx, fromCell.y + dy)
				if (dx == 0 and dy == 0) or simquery.isOpenExit(fromCell.exits[simquery.getDirectionFromDelta(dx, dy)]) then
					local exit = cell and cell.exits[dir]
					if exit and exit.door and simquery.isClosedDoor(exit) ~= true
					        and not (exit.closed == nil and exit.locked) -- exit is not broken
					        -- Don't ever slam the following doors
					        and exit.keybits ~= simdefs.DOOR_KEYS.ELEVATOR_INUSE
					        and exit.keybits ~= simdefs.DOOR_KEYS.ELEVATOR
					        and exit.keybits ~= simdefs.DOOR_KEYS.FINAL_LEVEL
					        and exit.keybits ~= simdefs.DOOR_KEYS.GUARD
					then
						table.insert(exits, {x = cell.x, y = cell.y, dir = dir})
					end
				end
			end

		end

		return targets.exitTarget(game, exits, self, abilityOwner, abilityUser)
	end,

	-- Is the agent able to slam the door?
	canUseAbility = function (self, sim, unit, abilityUser, x, y, dir) -- 'unit' should be the agent that has this ability. abilityUser should be the agent to right?
		-- Agents should always be able to slam an open door if they have enought AP
		if unit:getMP() < simdefs.EAA_DOOR_SLAM_MP_COST then
			return false, string.format(STRINGS.UI.REASON.REQUIRES_AP, simdefs.EAA_DOOR_SLAM_MP_COST)
		end
		if x and y then
			local isDoorClosed = simquery["isClosedDoor"] -- Have to get the function from simquery -- KLEI: Sooo brutal, but due to circular includes, itemdefs cannot directly reference simquery functions
			local cell = sim:getCell(x, y)
			local exit = cell.exits[dir]
			if isDoorClosed(exit) then
				return false, STRINGS.EXTRA_AGENT_ABILITIES.TOOLTIPS.DOOR_SLAM_REASON_DOOR_CLOSED
			end
			if exit.closed == nil and exit.locked then
				return false, STRINGS.UI.DOORS.BROKEN
			end
		end
		if unit ~= abilityUser then
			return false
		end
		return true
	end,

	-- Slam the door, leaving the door closed and creating a noise
	executeAbility = function (self, sim, unit, userUnit, x0, y0, dir)
		-- Charge unit MP
		userUnit:useMP(simdefs.EAA_DOOR_SLAM_MP_COST, sim)
		-- Generate a sound. Must do this first because 'simactions.userDoorAction()' pauses the game until the door is closed
		sim:emitSound({path = "SpySociety/Actions/door_close", range = 5}, x0, y0, userUnit, {{ x = x0, y = y0 }})
		sim:dispatchEvent(simdefs.EV_PLAY_SOUND, "SpySociety/Actions/door_close")
		-- Close the door
		simactions.useDoorAction(sim, simdefs.EXITOP_CLOSE, userUnit:getID(), x0, y0, dir)
	end

}
return doorSlam
