----------------------------------------------------------------------
-- EXTRA AGENT ABILITIES for Invisible Inc.
--
-- This file is contains functions to modify guard behavior.
-- These functions allow guards to see items and interact with them,
-- see strings.lua or modinit.lua or this file for more information.

----------------------------------------------------------------------
-- includes go here:
local actions = include("sim/btree/actions")
local inventory = include("sim/inventory")
local senses = include("sim/btree/senses")
local simdefs = include("sim/simdefs")
local util = include("modules/util")

----------------------------------------------------------------------
-- Constants
-- These values are also used by modinit to setup generation options
local ITEM_SIGHTABLE_VALUES =
{
	OFF = 0, -- Vanilla behavior
	PARTIAL = 1, -- Guards can only see items in direct vision
	FULL = 2 -- Guards can see items in direct and periphal vision
}
local GUARDS_ITEM_RESPONSE_VALUES =
{
	INVESTIGATE = 0,
	MIX = 1,
	ALERT = 2
}
local ITEM_DESTROY_VALUES =
{
	VANILLA = 0, -- Vanilla behavior
	PICKUP = 1, -- Guards pickup items if they have space, else they ignore
	PICKUP_DESTROY = 2, -- Guards pickup items if they have space, else they destory
	DESTROY = 3 -- Guards always destroy items they find on the ground
}

-- The following variables store information about what campaign generation options the user choose.
-- Once the campaign starts, these are effectively constants.

-- Information about our mod
local ITEM_SIGHTABLE = ITEM_SIGHTABLE_VALUES.FULL
local ITEM_DESTROY = ITEM_DESTROY_VALUES.PICKUP
local GUARDS_LOOT_CORPSES = false
-- local GUARDS_ITEM_RESPONSE = GUARDS_ITEM_RESPONSE_VALUES.ALERT -- value from campaign generations not needed because of ITEM_RESPONSE_TABLE
-- Information about other mods

-- Table of item responses, needed when GUARDS_ITEM_RESPONSE is set to MIX
local ITEM_RESPONSE_TABLE = {}

-- Function called by modInit that informs this file what campaign generation options where choosen by the user
local function setModInfo(modApi, dataPath, optionSightable, optionItemResponse, optionDestroy, optionGuardsLootCorpses)
	-- Get information about our mod
	ITEM_SIGHTABLE = optionSightable
	ITEM_DESTROY = optionDestroy
	GUARDS_LOOT_CORPSES = optionGuardsLootCorpses
	if optionItemResponse == GUARDS_ITEM_RESPONSE_VALUES.INVESTIGATE then
		ITEM_RESPONSE_TABLE = {["_DEFAULT_RESPONSE"] = "investigate"}
	elseif optionItemResponse == GUARDS_ITEM_RESPONSE_VALUES.ALERT then
		ITEM_RESPONSE_TABLE = {["_DEFAULT_RESPONSE"] = "alert"}
	elseif optionItemResponse == GUARDS_ITEM_RESPONSE_VALUES.MIX then
		ITEM_RESPONSE_TABLE = dofile(dataPath.."/custom_item_response.lua")
	else
		log:write("Unknown value of "..util.stringize(optionItemResponse, 1).." encountered for Guards Item Response.")
		ITEM_RESPONSE_TABLE = {["_DEFAULT_RESPONSE"] = "alert"}
	end
end
-- Called by modinit when the mod is completely unloaded.
-- We need to disable all of our changes as best we can in this function
local function unload()
	ITEM_SIGHTABLE = ITEM_SIGHTABLE_VALUES.OFF
	ITEM_RESPONSE_TABLE = {["_DEFAULT_RESPONSE"] = "none"}
	ITEM_DESTROY = ITEM_DESTROY_VALUES.VANILLA
	GUARDS_LOOT_CORPSES = false
end


----------------------------------------------------------------------
-- "sim/btree/actions" helper functions:

-- guardCanCarry returns false if the items is carried by another agent/guard
-- Since we want to be able to loot downed agents, we need a custom canCarry function
local function guardCanCarry(sim, itemUnit, guardUnit)
	-- log:write("Extra Agent Abilities > guardCanCarry called() - item:|"..tostring(itemUnit:getUnitData().name).."|")
	if not inventory.isCarryable(itemUnit) then
		-- log:write("Extra Agent Abilities >  cannot carry because item does not have carryable ability")
		return false
	end
	if itemUnit:getTraits().cantdrop == true then
		-- log:write("Extra Agent Abilities >  cannot carry because of trait cantdrop")
		return false
	end
	if guardUnit:getTraits().isDrone == true then
		-- log:write("Extra Agent Abilities >  cannot carry because of trait isDrone")
		return false
	end
	if guardUnit:getInventoryCount() >= (guardUnit:getTraits().inventoryMaxSize or 0) then
		-- log:write("Extra Agent Abilities >  cannot carry because of inventory space. Count:|"..tostring(guardUnit:getInventoryCount()).."| MaxSize:|"..tostring((guardUnit:getTraits().inventoryMaxSize or 0)).."|")
		return false
	end
	if itemUnit:getTraits().augment == true and itemUnit:getTraits().installed == true then
		-- log:write("Extra Agent Abilities >  cannot carry installed augments")
		return false -- Gotta say though, it was hilarious to have augments ripped out of agents brains before I remembered to check if the augment was installed.
	end
	if itemUnit:getTraits().impass == true or itemUnit:getTraits().dynamicImpass == true then
		-- log:write("Extra Agent Abilities >  cannot carry because of impass")
		return false
	end
	if itemUnit:getTraits().explodes ~= nil and itemUnit:getTraits().deployed == true then
		-- log:write("Extra Agent Abilities >  cannot carry because it's a live grenade")
		return false
	end
	-- log:write("Extra Agent Abilities >  returning true, guardCanCarry "..tostring(itemUnit:getUnitData().name))
	return true
end

-- Returns true if the item should not be destroyed under any circumstances
-- Returns false if the items is safe to destroy
local function doNotDestroy(itemUnit)
	-- log:write("Extra Agent Abilities > doNotDestroy called() - item:|"..tostring(itemUnit:getUnitData().name).."|")
	if itemUnit:getTraits().eaa_doNotDestroy == true then
		-- It's a mission critical item.
		-- log:write("Extra Agent Abilities >  mission critical item, returning true")
		return true
	elseif itemUnit:getTraits().installed == true then
		-- It's an installed augment.
		-- log:write("Extra Agent Abilities >  installed augment, returning true")
		return true
	elseif itemUnit:getTraits().explodes ~= nil and itemUnit:getTraits().deployed == true then
		-- It's a live grenade! Trying to destroy will cause a game crash when it tries to blow up next turn
		-- log:write("Extra Agent Abilities >  live grenade, returning true")
		return true
	end
	-- log:write("Extra Agent Abilities >  item safe for destroying. Returning false")
	return false
end

-- Custom functions used instead of inventory functions due to graphical issues
-- and problems involving units cannot be warped from nil to nil
local function ripItemOutOfAgentsInventory(agentUnit, itemUnit)
	agentUnit:removeChild(itemUnit)
	if itemUnit:getTraits().drop_dropdisguise then
		agentUnit:setDisguise(false)
	end
	if itemUnit:getTraits().equipped then
		itemUnit:getTraits().equipped = nil
	end
end

local function pickupItemOutOfAgentsInventory(sim, guardUnit, agentUnit, itemUnit)
	ripItemOutOfAgentsInventory(agentUnit, itemUnit)
	if itemUnit:getTraits().pickup_trigger then
		sim:triggerEvent(itemUnit:getTraits().pickup_trigger)
	end
	guardUnit._sim:dispatchEvent(simdefs.EV_UNIT_PICKEDUP, {unit=itemUnit})
	sim:triggerEvent(simdefs.TRG_UNIT_PICKEDUP, {item=itemUnit, unit=guardUnit})
	guardUnit:addChild(itemUnit)
	sim:triggerEvent("agentGotItem", {item=itemUnit})
	guardUnit._sim:dispatchEvent(simdefs.EV_PLAY_SOUND, "SpySociety/HUD/gameplay/HUD_ItemStorage_PutIn")
	guardUnit._sim:dispatchEvent(simdefs.EV_UNIT_GET_ITEM, {item = itemUnit})
end

local function destroyItemOutOfAgentsInventory(sim, agentUnit, itemUnit)
	ripItemOutOfAgentsInventory(agentUnit, itemUnit)
	sim:despawnUnit(itemUnit)
end

-- Flags an item as seen so guards don't reinvestigate it.
local function markItemSeen(guardID, item)
	if item._eaa_investigatedBy == nil then
		item._eaa_investigatedBy = {[guardID] = true}
	else
		item._eaa_investigatedBy[guardID] = true
	end
end

local function pickupItem(sim, guardUnit, agentUnit, itemUnit)
	if agentUnit ~= nil then
		pickupItemOutOfAgentsInventory(sim, guardUnit, agentUnit, itemUnit)
	else
		inventory.pickupItem(sim, guardUnit, itemUnit)
	end
	sim:dispatchEvent(simdefs.EV_UNIT_PICKUP, {unitID = guardUnit:getID()})
	guardUnit:resetAllAiming()
	sim:dispatchEvent(simdefs.EV_UNIT_REFRESH, {unit = guardUnit})
end

local function destroyItem(sim, agentUnit, itemUnit)
	if agentUnit ~= nil then
		destroyItemOutOfAgentsInventory(sim, agentUnit, itemUnit)
	else
		sim:warpUnit(itemUnit)
		sim:despawnUnit(itemUnit)
	end
end

-- handleUnit controls whether or not a guard picks up an item, or destroys it
-- It returns true if the item is destroyed, and false in all other cases
local function handleUnit(sim, itemUnit, guardUnit, agentUnit)
	if ITEM_DESTROY == ITEM_DESTROY_VALUES.DESTROY then
		-- Destroy the item no matter what (unless it's an installed augment), but don't harm precious items (like incognita)
		if doNotDestroy(itemUnit) == false then
			destroyItem(sim, agentUnit, itemUnit)
			return true
		else
			markItemSeen(guardUnit:getID(), itemUnit)
			return false
		end
	elseif ITEM_DESTROY == ITEM_DESTROY_VALUES.PICKUP_DESTROY then
		-- Only destroy the item if we cannot carry it (usually means our inventory is full)
		if guardCanCarry(sim, itemUnit, guardUnit) then
			pickupItem(sim, guardUnit, agentUnit, itemUnit)
			return false
		elseif doNotDestroy(itemUnit) == false then
			destroyItem(sim, agentUnit, itemUnit)
			return true
		else
			-- Don't repeatedly investigate the same item (unless it moves)
			markItemSeen(guardUnit:getID(), itemUnit)
			return false
		end
	elseif ITEM_DESTROY == ITEM_DESTROY_VALUES.PICKUP then
		-- Never destroy an item. But do pick it up if we have space
		if guardCanCarry(sim, itemUnit, guardUnit) then
			pickupItem(sim, guardUnit, agentUnit, itemUnit)
			return false
		else
			markItemSeen(guardUnit:getID(), itemUnit)
		end
	else -- ITEM_DESTROY == ITEM_DESTROY_VALUES.VANILLA
		-- Vanilla behavior already occurs during the old_ function call. So we will just do almost nothing here.
		markItemSeen(guardUnit:getID(), itemUnit) -- still need to mark as seen to stop repeat investigations
		return false
	end
end

----------------------------------------------------------------------
-- "sim/btree/actions" overrides:
-- This function controls what happens when a guard reaches an 'interest' spot
-- We are using it to control pickup/destruction of items that a guard investigates
local old_MarkInterestInvestigated = actions.MarkInterestInvestigated
actions.MarkInterestInvestigated = function(sim, unit)
	if not unit:getBrain():getInterest() then
		return simdefs.BSTATE_FAILED
	end

	local cell = unit:getSim():getCell(unit:getLocation())
	local units = {}
	local downed_agents = {}
	for _,checkUnit in ipairs(cell.units) do
		if checkUnit ~= unit then
			if checkUnit:getTraits().isAgent and checkUnit:getBrain() == nil then
				table.insert(downed_agents, checkUnit)
			elseif inventory.canCarry(unit, checkUnit) then
				table.insert(units, checkUnit)
			elseif checkUnit:getTraits().sightable == true and checkUnit:getTraits().deployed == true then
				table.insert(units, checkUnit)
			end
		end
	end

	local destroyedUnit = false

	if #units > 0 and ITEM_SIGHTABLE ~= ITEM_SIGHTABLE_VALUES.OFF then
		for _,pickUpUnit in ipairs(units) do
			if handleUnit(sim, pickUpUnit, unit, nil) == true then
				destroyedUnit = true
			end
		end
	end
	if #downed_agents > 0 and GUARDS_LOOT_CORPSES == true then
		for _,agent in ipairs(downed_agents) do
			units = {}
			for j,checkUnit in ipairs(agent:getChildren()) do
				if inventory.canCarry(unit, checkUnit) then
					table.insert(units,checkUnit)
				end
			end
			if #units > 0 then
				for _,pickUpUnit in ipairs(units) do
					if handleUnit(sim, pickUpUnit, unit, agent) == true then
						destroyedUnit = true
					end
				end
			end
		end
	end

	if destroyedUnit then
		local params = {color ={{symbol="inner_line",r=1,g=0,b=0,a=0.75},{symbol="wall_digital",r=1,g=0,b=0,a=0.75},{symbol="boxy_tail",r=1,g=0,b=0,a=0.75},{symbol="boxy",r=1,g=0,b=0,a=0.75}} }
		sim:dispatchEvent( simdefs.EV_UNIT_ADD_FX, { unit = unit, kanim = "fx/emp_effect", symbol = "character", anim="idle", above=true, params=params} )
		local x1, y1 = unit:getLocation()
		sim:dispatchEvent( simdefs.EV_PLAY_SOUND, {sound="SpySociety/HitResponse/hitby_tazer_flesh", x=x1,y=y1} )
	end

	return old_MarkInterestInvestigated(sim, unit)
end

----------------------------------------------------------------------
-- "sim/inventory" overrides
local old_pickupItem = inventory.pickupItem
inventory.pickupItem = function(sim, unit, item)
	-- Reset who has already investigated this item
	item._eaa_investigatedBy = nil
	old_pickupItem(sim, unit, item)
end

----------------------------------------------------------------------
-- "sim/btree/senses" helper functions:

-- Returns true if the specified guard has not seen the item yet
local function itemAlreadySeen(guardID, item)
	if item._eaa_investigatedBy == nil then
		return false
	else
		return (item._eaa_investigatedBy[guardID] == true) -- '==' operator ensures we return true/false and not nil
	end
end

local function processItemResponseLevel(guardSenses, x, y, item, senseSight, responseOverride)
	-- Don't respond to items we have already seen.
	if not itemAlreadySeen(guardSenses.unit:getID(), item) then
		local response = responseOverride or ITEM_RESPONSE_TABLE[util.toupper(item:getName())] or ITEM_RESPONSE_TABLE["_DEFAULT_RESPONSE"]
		if response == "alert" then
			guardSenses:addInterest(x, y, senseSight, simdefs.REASON_FOUNDOBJECT)
		elseif response == "investigate" then
			guardSenses:addInterest(x, y, senseSight, simdefs.REASON_NOTICED)
		end
	end
end

----------------------------------------------------------------------
-- "sim/btree/senses" overrides:

-- Handles when something appears in a guards vision
local old_processAppearedTrigger = senses.processAppearedTrigger
senses.processAppearedTrigger = function(self, sim, evData)
	old_processAppearedTrigger(self, sim, evData)
	if ITEM_SIGHTABLE ~= ITEM_SIGHTABLE_VALUES.OFF then
		local seerUnit = sim:getUnit(evData.seerID)
		if seerUnit ~= self.unit then
			return 	--the seer isn't us
		end

		if self.unit:isKO() or self.unit:isDead() then
			return
		end

		local x,y = evData.unit:getLocation()

		if evData.unit.hasAbility and evData.unit:hasAbility("carryable") and evData.unit:getTraits().sightable then
			processItemResponseLevel(self, x, y, evData.unit, simdefs.SENSE_SIGHT)
		end
	end
end
-- Handles when a guards vision moves
local old_checkPeripheralVision = senses.checkPeripheralVision
senses.checkPeripheralVision = function(self)
	old_checkPeripheralVision(self)
	if ITEM_SIGHTABLE == ITEM_SIGHTABLE_VALUES.FULL then
		local losCoords, cells = {}, {}
		local sim = self.unit:getSim()
		sim:getLOS():getPeripheralVizCells( self.unit:getID(), losCoords )
		for i = 1, #losCoords, 2 do
			local x, y = losCoords[i], losCoords[i+1]
			table.insert(cells, sim:getCell(x, y))
		end
		for _, cell in ipairs(cells) do
			for _, unit in ipairs(cell.units) do
				local canSee, canSense = sim:canUnitSeeUnit(self.unit, unit)
				if not canSee and canSense then
					if unit.hasAbility and unit:hasAbility("carryable") and unit:getTraits().sightable then
						processItemResponseLevel(self, cell.x, cell.y, unit, simdefs.SENSE_PERIPHERAL, "investigate")
					end
				end
			end
		end
	end
end
-- Handles when a something warps into a guards vision.
local old_processWarpTrigger = senses.processWarpTrigger
senses.processWarpTrigger = function(self, sim, evData)
	old_processWarpTrigger(self, sim, evData)
	if ITEM_SIGHTABLE == ITEM_SIGHTABLE_VALUES.FULL then
		if evData.unit:isValid() and evData.unit.hasAbility and evData.unit:hasAbility("carryable") and evData.unit:getTraits().sightable then -- check if this is an item
			if evData.to_cell then -- check to see if the item warped to a location (if it warped to nil we don't care because we know we cannot see it anymore)
				local canSee, canSense = sim:canUnitSeeUnit(self.unit, evData.unit)
				if canSee == true or canSense == true then
					if not canSee and canSense then
						processItemResponseLevel(self, evData.to_cell.x, evData.to_cell.y, evData.unit, simdefs.SENSE_PERIPHERAL, "investigate")
					elseif canSee then
						processItemResponseLevel(self, evData.to_cell.x, evData.to_cell.y, evData.unit, simdefs.SENSE_SIGHT)
					end
				end
			end
		end
	end
end

----------------------------------------------------------------------
return {
	ITEM_SIGHTABLE_VALUES = ITEM_SIGHTABLE_VALUES,
	GUARDS_ITEM_RESPONSE_VALUES = GUARDS_ITEM_RESPONSE_VALUES,
	ITEM_DESTROY_VALUES = ITEM_DESTROY_VALUES,
	setModInfo = setModInfo,
	unload = unload
}
