----------------------------------------
-- This ability allows normal items to be thrown

local abilityutil = include("sim/abilities/abilityutil")
local inventory = include("sim/inventory")
local simdefs = include("sim/simdefs" )
local simquery = include("sim/simquery")

local throwShared = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/throwShared")

local function itemThrowWarpUnit(sim, grenadeUnit, cell)
	if grenadeUnit:getUnitData().type == "simemppack" then
		-- This is a hack to stop EMP packs from generating an error when calling warpUnit(), as they add a trigger when they are warped, and in the process of dropItem(), grenadeUnit has already been warped once.
		sim:removeTrigger(simdefs.TRG_END_TURN, grenadeUnit)
	end
	sim:warpUnit(grenadeUnit, cell)
end

local itemThrow =
{
	name = STRINGS.ABILITIES.THROW,

	getName = function(self, sim, unit, userUnit)
		return self.name
	end,

	createToolTip = function(self, sim, unit, targetCell)
		return abilityutil.formatToolTip(STRINGS.ABILITIES.THROW,  STRINGS.ABILITIES.THROW_DESC, 1)
	end,

	profile_icon = "gui/icons/action_icons/Action_icon_Small/icon-item_shoot_small.png",
	usesMP = true,

	-- Changed acquireTargets to always set range to zero. This stops squares from being highlighted when previewing throws for items that normally would affect things in a range, such as EMP packs
	acquireTargets = function(self, targets, game, sim, grenadeUnit, unit)
		if grenadeUnit:getTraits() == nil or grenadeUnit:getTraits().eaa_isThrowing ~= true then
			return nil
		end
		if not self:canUseAbility(sim, grenadeUnit, unit) then
			return nil
		end
		return throwShared.itemThrowTarget(game, 0, sim, unit, unit:getTraits().maxThrow, grenadeUnit:getTraits().targeting_ignoreLOS, throwShared.modInfo.ITEM_THROW_MP_COST)
	end,
	-- Changed canUseAbility to remove item MP, PWR, and cooldown checks
	canUseAbility = function(self, sim, grenadeUnit, unit, targetCell)
		if grenadeUnit:getTraits() == nil or grenadeUnit:getTraits().eaa_isThrowing ~= true then
			return false
		end
		if unit:getMP() < throwShared.modInfo.ITEM_THROW_MP_COST then
			return false, string.format(STRINGS.UI.REASON.REQUIRES_AP, throwShared.modInfo.ITEM_THROW_MP_COST)
		end

		if targetCell then
			local targetX,targetY = unpack(targetCell)
			local unitX, unitY = unit:getLocation()
			local raycastX, raycastY = sim:getLOS():raycast(unitX, unitY, targetX, targetY)
			if raycastX ~= targetX or raycastY ~= targetY then
				return false
			end
		end

		return true
	end,

	-- Changed executeAbility: removed references to NPC usage, added some code to get ride of errors when throwing items, and to allow the item to be visible when thrown
	executeAbility = function(self, sim, grenadeUnit, userUnit, targetCell)
		-- local sim = grenadeUnit:getSim()
		local current_x,current_y = userUnit:getLocation()
		local target_x,target_y = unpack(targetCell)
		userUnit:useMP(throwShared.modInfo.ITEM_THROW_MP_COST, sim)
		userUnit:getTraits().throwing = true

		local facing = simquery.getDirectionFromDelta(target_x - current_x, target_y - current_y)
		simquery.suggestAgentFacing(userUnit, facing)

		if userUnit:isValid() and not userUnit:getTraits().interrupted then
			sim:dispatchEvent(simdefs.EV_UNIT_THROW, {unit = userUnit, x1 = target_x, y1 = target_y, facing = facing})
			inventory.dropItem(sim, userUnit, grenadeUnit)

			if sim:isVersion("0.17.15") and userUnit:getTraits().data_hacking then
				userUnit:stopHacking(sim)
			end

			sim:dispatchEvent(simdefs.EV_UNIT_THROWN, {unit = grenadeUnit, x = target_x, y = target_y})

			if current_x ~= target_x or current_y ~= target_y then
				if throwShared.modInfo.GUARDS_SEE_THROW == true then
					-- log:write(" start("..tostring(current_x)..","..tostring(current_y)..")")
					-- log:write("target("..tostring( target_x)..","..tostring( target_y)..")")
					-- warp the item one square at a time so guards can see it fly through the air
					for _, point in ipairs(throwShared.getThrowTravelPoints({x = current_x, y = current_y}, {x = target_x, y = target_y})) do
						-- log:write(">     point("..tostring(point.x)..","..tostring(point.y)..")")
						itemThrowWarpUnit(sim, grenadeUnit, sim:getCell(point.x, point.y))
					end
				else
					itemThrowWarpUnit(sim, grenadeUnit, sim:getCell(target_x, target_y))
				end
			end

			sim:dispatchEvent(simdefs.EV_UNIT_STOP_THROW, {unitID = userUnit:getID(), x1 = target_x, y1 = target_y, facing = facing})

			sim:processReactions(userUnit)
		end
		userUnit:getTraits().throwing = nil
		if userUnit:isValid() and not userUnit:getTraits().interrupted then
			simquery.suggestAgentFacing(userUnit, facing)
		end

		-- Remove the throw ability so the player must go through the selector to choose to throw
		grenadeUnit:getTraits().eaa_isThrowing = nil
		sim:triggerEvent("EAA_TRG_ITEM_THROW_CANCEL")
	end
}

return itemThrow
