----------------------------------------------------------------------
-- throwExtender.lua
--
-- Contains functions to extend a throw ability.
-- Used to allow 'Guards See Item Throws' to work with other implementations of throw.
local abilitydefs = include("sim/abilitydefs")
local simfactory = include("sim/simfactory")
local util = include("modules/util")

local throwShared = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/throwShared")

-- Table of extended versions of ability functions
local extendedFunctions = {
	acquireTargets = function(self, targets, game, sim, grenadeUnit, unit)
		local target = self:eaa_old_acquireTargets(targets, game, sim, grenadeUnit, unit)
		if target ~= nil then
			return throwShared.itemThrowTarget(target._game, target.range, target.sim, target.unit, target.unitRange, target._ignoreLOS, throwShared.modInfo.GRENADE_THROW_MP_COST)
		else
			return nil
		end
	end,

	canUseAbility = function(self, sim, grenadeUnit, unit, targetCell)
		-- Only agents get charged MP, do not charge guards.
		if unit:getBrain() == nil and unit:getMP() < throwShared.modInfo.GRENADE_THROW_MP_COST then
			return false, string.format(STRINGS.UI.REASON.REQUIRES_AP, throwShared.modInfo.GRENADE_THROW_MP_COST)
		end
		return self:eaa_old_canUseAbility(sim, grenadeUnit, unit, targetCell)
	end,

	executeAbility = function(self, sim, grenadeUnit, userUnit, targetCell)
		-- Only agents get charged MP, do not charge guards.
		if userUnit:getBrain() == nil then
			userUnit:useMP(throwShared.modInfo.GRENADE_THROW_MP_COST, sim)
		end
		if throwShared.modInfo.GUARDS_SEE_THROW == true then
			-- spawn a dummy (since eaa_old_executeAbility despawns the original for consumable grenades)
			local dummyUnitData = util.extend(grenadeUnit:getUnitData()){}
			local dummyUnit = simfactory.createUnit(dummyUnitData, sim)
			sim:spawnUnit(dummyUnit)
			if dummyUnit:getTraits().holoProjector == true then
				-- Make the throw visible, while the actual holoProjector remains invisible once placed
				dummyUnit:getTraits().sightable = true
				dummyUnit:getTraits().hidesInCover = true
			end
			-- warp the item one square at a time so guards can see it fly through the air
			local start_x, start_y = userUnit:getLocation()
			local target_x, target_y = unpack(targetCell)
			-- log:write(" start("..tostring( start_x)..","..tostring( start_y)..")")
			-- log:write("target("..tostring(target_x)..","..tostring(target_y)..")")
			for _, point in ipairs(throwShared.getThrowTravelPoints({x = start_x, y = start_y}, {x = target_x, y = target_y})) do
				-- log:write(">     point("..tostring(point.x)..","..tostring(point.y)..")")
				sim:warpUnit(dummyUnit, sim:getCell(point.x, point.y))
			end
			-- Remove dummy from world
			if dummyUnit:getLocation() ~= nil then
				sim:warpUnit(dummyUnit)
			end
			sim:despawnUnit(dummyUnit)
		end
		self:eaa_old_executeAbility(sim, grenadeUnit, userUnit, targetCell)
	end
}

local function extendThrow(abilityName)
	local throwAbility = abilitydefs.lookupAbility(abilityName)
	if throwAbility then
		if throwAbility.eaa_old_acquireTargets == nil then
			throwAbility.eaa_old_acquireTargets = throwAbility.acquireTargets
			throwAbility.acquireTargets = extendedFunctions.acquireTargets
		end
		if throwAbility.eaa_old_canUseAbility == nil then
			throwAbility.eaa_old_canUseAbility = throwAbility.canUseAbility
			throwAbility.canUseAbility = extendedFunctions.canUseAbility
		end
		if throwAbility.eaa_old_executeAbility == nil then
			throwAbility.eaa_old_executeAbility = throwAbility.executeAbility
			throwAbility.executeAbility = extendedFunctions.executeAbility
		end
	end
end

-- Extend Throws
-- Vanilla throws
extendThrow("throw")
-- Talon Recruitment throws
extendThrow("sombra_throw_translocator")
extendThrow("widow_throw_mine")
-- Agent Mod Combo Throws
extendThrow("throw_2")
extendThrow("throw_decoy_2")
extendThrow("throw_emp_2")
extendThrow("throw_spider_2")
-- Flavorful Agents throws
extendThrow("banks_throwParalyze")
extendThrow("throw_snap")
-- Assassin throw
--extendThrow("eagleSight") -- Don't extend this one, it's supposed to be invisible anyway (extending it makes the throw visible, though of course the drone is still invisible)

-- Ramblings on mod compatibility:
-- We do not have full compatibility with Flavorful Agent's banks paralyzer or Gunpoint's bullfrog.
-- They should still work but will not display the 'deploy' button (tosses will have to be canceled with right click).

return nil
