----------------------------------------
-- Contains code shared by itemThrow.lua and throw.lua

local cdefs = include("client_defs")
local util = include("modules/util")
local abilitydefs = include("sim/abilitydefs")
local simdefs = include( "sim/simdefs" )
local simquery = include("sim/simquery")
local targeting = include( "hud/targeting")

----------------------------------------
-- mod options, stored in a table so that the values are passed by reference instead of value
local modInfo = {
	ITEM_THROW_MP_COST = 1,
	GRENADE_THROW_MP_COST = 1,
	GUARDS_SEE_THROW = true
}

local ICON_COLOR_INACTIVE = {
	FREE = util.color(126/255, 128/255, 60/255),
	ACTION = util.color(70/255, 128/255, 128/255)
}
----------------------------------------
-- Functions for use by modinit
local function setModInfo(itemThrowMPCost, grenadeThrowMPCost, guardsSeeThrow, itemSightable)
	modInfo.ITEM_THROW_MP_COST = itemThrowMPCost
	modInfo.GRENADE_THROW_MP_COST = grenadeThrowMPCost
	modInfo.GUARDS_SEE_THROW = guardsSeeThrow and itemSightable -- both options must be on for guards to see throws
	local selectItemToss = abilitydefs.lookupAbility("selectItemToss")
	local selectItemDeploy = abilitydefs.lookupAbility("selectItemDeploy")
	if selectItemToss ~= nil then
		if modInfo.ITEM_THROW_MP_COST > 0 then
			selectItemToss.usesMP = true
			selectItemToss.ICON_COLOR_ACTIVE = cdefs.COLOR_ACTION
			selectItemToss.ICON_COLOR_INACTIVE = ICON_COLOR_INACTIVE.ACTION
			selectItemToss.iconColor = selectItemToss.ICON_COLOR_INACTIVE
		else
			selectItemToss.usesMP = false
			selectItemToss.ICON_COLOR_ACTIVE = cdefs.COLOR_FREE
			selectItemToss.ICON_COLOR_INACTIVE = ICON_COLOR_INACTIVE.FREE
			selectItemToss.iconColor = selectItemToss.ICON_COLOR_INACTIVE
		end
	end
	if selectItemDeploy ~= nil then
		if modInfo.GRENADE_THROW_MP_COST > 0 then
			selectItemDeploy.usesMP = true
			selectItemDeploy.ICON_COLOR_ACTIVE = cdefs.COLOR_ACTION
			selectItemDeploy.ICON_COLOR_INACTIVE = ICON_COLOR_INACTIVE.ACTION
			selectItemDeploy.iconColor = selectItemDeploy.ICON_COLOR_ACTIVE
		else
			selectItemDeploy.usesMP = false
			selectItemDeploy.ICON_COLOR_ACTIVE = cdefs.COLOR_FREE
			selectItemDeploy.ICON_COLOR_INACTIVE = ICON_COLOR_INACTIVE.FREE
			selectItemDeploy.iconColor = selectItemDeploy.ICON_COLOR_ACTIVE
		end
	end
end

local function unload()
	modInfo.ITEM_THROW_MP_COST = 0
	modInfo.GRENADE_THROW_MP_COST = 0
	modInfo.GUARDS_SEE_THROW = false
end

----------------------------------------
-- Functions for general use

-- Calculates the points a thrown item will pass through and returns them in an orderd list
-- start -> table of the form {x, y}
-- stop  -> table of the form {x, y}
-- return -> table of the form {{x1, y1}, {x2, y2}, ... , {xn, yn}}
local function getThrowTravelPoints(start, stop)
	local points = {}

	-- warp the item one square at a time so guards can see it fly through the air
	local real_loc = {x = start.x, y = start.y} -- the items location expressed as a real (decimal) number
	local int_loc  = {x = start.x, y = start.y} -- the items location expressed as an intenger
	-- Calculate how much to move each increment
	local move = {x = nil, y = nil} -- How much to move real_loc each increment
	-- Check if we only need to move in one dimension
	if start.x == stop.x then
		move.x = 0
		if start.y > stop.y then
			move.y = -1
		else
			move.y = 1
		end
	elseif start.y == stop.y then
		move.y = 0
		if start.x > stop.x then
			move.x = -1
		else
			move.x = 1
		end
	else
		-- Calculate slope (into its x and y components)
		local delta = {
			x = stop.x - start.x,
			y = stop.y - start.y
		}
		-- Modify the slope so that the largest value equals 1
		if math.abs(delta.x) > math.abs(delta.y) then
			move.x = delta.x > 0 and 1 or -1
			move.y = delta.y / math.abs(delta.x)
		else
			move.x = delta.x / math.abs(delta.y)
			move.y = delta.y > 0 and 1 or -1
		end
	end
	-- Move along the vector in 1 unit increments
	while int_loc.x ~= stop.x or int_loc.y ~= stop.y do
		real_loc.x = real_loc.x + move.x
		real_loc.y = real_loc.y + move.y
		int_loc.x = math.floor(real_loc.x + 0.5) -- adding 0.5 causes the number to be rounded
		int_loc.y = math.floor(real_loc.y + 0.5)
		table.insert(points, {x = int_loc.x, y = int_loc.y})
		-- Don't move past the target
		if int_loc.x == stop.x then
			move.x = 0
		end
		if int_loc.y == stop.y then
			move.y = 0
		end
	end
	return points
end

----------------------------------------
-- Classes

-- Modified target to allow AP preview (note this is the same code found in this mods throw.lua, if you change one spot, please change the other as well)
local itemThrowTarget = class(targeting.throwTarget)

function itemThrowTarget:init(game, range, sim, unit, isTargetFn, ignoreLOS, throwMPCost)
	targeting.throwTarget.init(self, game, range, sim, unit, isTargetFn, ignoreLOS)
	-- color scheme from board
	-- self._travelWatchedHiliteClr = util.color(255/255, 1/255, 1/255, 0.5)
	-- self._travelNoticedHiliteClr = util.color(100/255, 1/255, 1/255, 0.5)
	-- self._travelHiddenHiliteClr = util.color(88/255, 83/255, 44/255, 0.5)
	-- self._travelDefaultHiliteClr = util.color(0.5, 0.5, 0.5, 0.5)
	-- color scheme from agent walking
	-- self._travelWatchedHiliteClr = util.color(184/255, 13/255, 13/255, 0.5)
	-- self._travelNoticedHiliteClr = util.color(184/255, 184/255, 13/255, 0.5)
	-- self._travelHiddenHiliteClr = util.color(13/255, 184/255, 184/255, 0.5)
	-- self._travelDefaultHiliteClr = util.color(0.5, 0.5, 0.5, 0.5)
	-- even darker scheme
	self._travelWatchedHiliteClr = util.color(119/255, 8/255, 8/255, 0.5)
	self._travelNoticedHiliteClr = util.color(119/255, 119/255, 8/255, 0.5)
	self._travelHiddenHiliteClr = util.color(8/255, 119/255, 119/255, 0.5)
	self._travelDefaultHiliteClr = util.color(0.5, 0.5, 0.5, 0.5)
	self._game = game
	self._unit = unit
	self._throwMPCost = throwMPCost
end

-- Searches for a specified x, y coordinate in a table of {x,y} points
local function findPoint(x, y, points)
	for i, point in ipairs(points) do
		if x == point.x and y == point.y then
			return i
		end
	end
	return nil
end

function itemThrowTarget:onDraw()
	if self.mx and self.my and self:isValidTargetLoc(self.mx, self.my) then
		self._game.hud:previewAbilityAP(self._unit, self._throwMPCost)
		if modInfo.GUARDS_SEE_THROW == true then
			-- highlight the cells the item will fly through that guards might see
			local unitCell_x, unitCell_y = self._unit:getLocation()
			local travel_points = getThrowTravelPoints({x = unitCell_x, y = unitCell_y}, {x = self.mx, y = self.my})
			for _, point in ipairs(travel_points) do
				-- Determine color based on guard observation
				local cell_state = simquery.isCellWatched(self._unit._sim, self._game:getLocalPlayer(), point.x, point.y)
				if cell_state == simdefs.CELL_WATCHED then
					MOAIGfxDevice.setPenColor(self._travelWatchedHiliteClr:unpack())
				elseif cell_state == simdefs.CELL_NOTICED then
					MOAIGfxDevice.setPenColor(self._travelNoticedHiliteClr:unpack())
				else--if cell_state == simdefs.CELL_HIDDEN then
					MOAIGfxDevice.setPenColor(self._travelHiddenHiliteClr:unpack())
				-- else
				-- 	MOAIGfxDevice.setPenColor(self._travelDefaultHiliteClr:unpack())
				end
				local x0, y0 = self._game:cellToWorld(point.x + 0.3, point.y + 0.3)
				local x1, y1 = self._game:cellToWorld(point.x - 0.3, point.y - 0.3)
				MOAIDraw.fillRect(x0, y0, x1, y1)
			end
			-- highlight the cells that the grenade/item will impact when it explodes
			-- NOTE: may need to try to ommit the cells that we already highlighted...
			MOAIGfxDevice.setPenColor(unpack(self._hiliteClr))
			local coordinates = simquery.rasterCircle(self.sim, self.mx, self.my, self.range)
			for i = 1, #coordinates, 2 do
				local point = {x=coordinates[i], y=coordinates[i+1]}
				if findPoint(point.x, point.y, travel_points) == nil then
					local raycastX, raycastY = self.sim:getLOS():raycast(self.mx, self.my, point.x, point.y)
					if ((raycastX == point.x and raycastY == point.y) or self._ignoreLOS ) and self.sim:canPlayerSee(self.unit:getPlayerOwner(), point.x, point.y) then
						local x0, y0 = self._game:cellToWorld(point.x + 0.4, point.y + 0.4)
						local x1, y1 = self._game:cellToWorld(point.x - 0.4, point.y - 0.4)
						MOAIDraw.fillRect( x0, y0, x1, y1 )
					end
				else
					local x0, y0 = self._game:cellToWorld(point.x + 0.4, point.y + 0.4)
					local x1, y1 = self._game:cellToWorld(point.x + 0.3, point.y - 0.4)
					MOAIDraw.fillRect( x0, y0, x1, y1 )
					x0, y0 = self._game:cellToWorld(point.x + 0.3, point.y - 0.3)
					x1, y1 = self._game:cellToWorld(point.x - 0.4, point.y - 0.4)
					MOAIDraw.fillRect( x0, y0, x1, y1 )
					x0, y0 = self._game:cellToWorld(point.x - 0.3, point.y + 0.4)
					x1, y1 = self._game:cellToWorld(point.x - 0.4, point.y - 0.3)
					MOAIDraw.fillRect( x0, y0, x1, y1 )
					x0, y0 = self._game:cellToWorld(point.x + 0.3, point.y + 0.4)
					x1, y1 = self._game:cellToWorld(point.x - 0.3, point.y + 0.3)
					MOAIDraw.fillRect( x0, y0, x1, y1 )
				end
			end
		else
			return targeting.throwTarget.onDraw(self)
		end
		return true
	else
		self._game.hud:previewAbilityAP(self._unit, 0)
		return false
	end
end

function itemThrowTarget:endTargeting()
	targeting.throwTarget.endTargeting(self)
	self._game.hud:previewAbilityAP(self._unit, 0)
end

return {
	modInfo = modInfo,

	setModInfo = setModInfo,
	unload = unload,

	getThrowTravelPoints = getThrowTravelPoints,
	itemThrowTarget = itemThrowTarget
}
