----------------------------------------------------------------------
-- EXTRA AGENT ABILITIES for Invisible Inc.
--
-- This file is responsible for initializing the mod and telling the main program what's up



-- Important reminder to self: Include paths must be identical to the include path used by other lua files. Otherwise a duplicate version of the file will be loaded (and any changes made to the duplicate will not be carried to the original).

EXTRA_AGENT_ABILITIES_SCRIPT_PATH = "obvious bad file path is obvious" -- making this global because nobody left a good way outside of modinit to get to a mods scriptPath


local simdefs = include("sim/simdefs")
local util = include("modules/util")
-- Table of tables containing items modified by this mod
local modifiedItemsAndProps =
{
	addedSightable = {},
	addedHidesInCover = {}
}

----------------------------------------------------------------------
-- Local functions defs, defined before load(), unload(), etc.
--


-- LATE load/unload only. Adds doorSlam to all agents. Does not add doorSlam if the agent already has doorSlam
local function addDoorSlam()
	local agentdefs = include("sim/unitdefs/agentdefs")

	for name, agentdef in pairs(agentdefs) do
		-- Make sure we don't add the ability twice to the same agent
		if util.indexOf(agentdef.abilities, "doorSlam") == nil then
			table.insert(agentdef.abilities, "doorSlam")
		end
	end
end

-- LATE load/unload only. Searches for and removes the "doorSlam" from agents
-- Note, this will only remove the first instance of "doorSlam", if the agent got multiple copies of it the agent will still have "doorSlam" as an ability
local function removeDoorSlam()
	local removeIndex
	local agentdefs = include("sim/unitdefs/agentdefs")

	for name, agentdef in pairs(agentdefs) do
		removeIndex = util.indexOf( agentdef.abilities, "doorSlam")
		if removeIndex ~= nil then
			table.remove(agentdef.abilities, removeIndex)
		end
	end
	-- It would appear that at least one of the agent defs references commondefs (and not just a copy), so we need to make sure we remove doorSlam from commondefs as well
	local commondefs = include("sim/unitdefs/commondefs")
	removeIndex = util.indexOf(commondefs.DEFAULT_AGENT_ABILITIES, "doorSlam")
	if removeIndex ~= nil then
		table.remove(commondefs.DEFAULT_AGENT_ABILITIES, removeIndex)
	end
end

-- LATE load/unload only. Adds agentDoorBreak to all agents. Does not add agentDoorBreak if the agent already has agentDoorBreak
local function addAgentDoorBreak()
	local agentdefs = include("sim/unitdefs/agentdefs")

	for name, agentdef in pairs(agentdefs) do
		-- Make sure we don't add the ability twice to the same agent
		if util.indexOf(agentdef.abilities, "agentDoorBreak") == nil then
			agentdef.traits.doorBreakKOBonus = 1
			agentdef.traits.doorBreakArmorPiercingBonus = 0
			agentdef.traits.doorBreakFullImpactRange = 1
			agentdef.traits.doorBreakHalfImpactRange = 0
			table.insert(agentdef.abilities, "agentDoorBreak")
		end
	end
end

-- LATE load/unload only. Searches for and removes the "agentDoorBreak" from agents
-- Note, this will only remove the first instance of "agentDoorBreak", if the agent got multiple copies of it the agent will still have "agentDoorBreak" as an ability
local function removeAgentDoorBreak()
	local removeIndex
	local agentdefs = include("sim/unitdefs/agentdefs")

	for name, agentdef in pairs(agentdefs) do
		removeIndex = util.indexOf(agentdef.abilities, "agentDoorBreak")
		if removeIndex ~= nil then
			agentdef.traits.doorBreakKOBonus = nil
			agentdef.traits.doorBreakArmorPiercingBonus = nil
			agentdef.traits.doorBreakFullImpactRange = nil
			agentdef.traits.doorBreakHalfImpactRange = nil
			table.remove(agentdef.abilities, removeIndex)
		end
	end
	-- It would appear that at least one of the agent defs references commondefs (and not just a copy), so we need to make sure we remove agentDoorBreak from commondefs as well
	local commondefs = include("sim/unitdefs/commondefs")
	removeIndex = util.indexOf(commondefs.DEFAULT_AGENT_ABILITIES, "agentDoorBreak")
	if removeIndex ~= nil then
		table.remove(commondefs.DEFAULT_AGENT_ABILITIES, removeIndex)
	end
end

-- LATE load/unload only. Makes all items throwable. Items added to itemdefs after this function is called will not be made throwable
local function addItemThrow()
	local function isGrenade(unitdef)
		if (unitdef.rig and unitdef.rig == "grenaderig") or
		   (unitdef.type and string.find(unitdef.type, "grenade")) then
			return true
		end
		return false
	end

	local itemdefs = include("sim/unitdefs/itemdefs")
	local propdefs = include("sim/unitdefs/propdefs") -- Some items are props
	local itemrig = include("gameplay/itemrig") -- The rig that itemThrow items use

	local alreadyHasAbility

	for name, idef in pairs(itemdefs) do
		if idef.abilities ~= nil and util.indexOf(idef.abilities, "carryable") ~= nil then
			-- Check to make sure we don't add the same ability twice
			alreadyHasAbility = false
			for _, abilityName in pairs(idef.abilities) do
				if abilityName == "itemThrow" then
					alreadyHasAbility = true
					break
				end
			end
			if not alreadyHasAbility then
				table.insert(idef.abilities, "selectItemToss")
				if isGrenade(idef) then
					-- Only add the deploy option to grenades
					table.insert(idef.abilities, "selectItemDeploy")
				end
				table.insert(idef.abilities, "itemThrow")
			end
		end
	end

	for name, pdef in pairs(propdefs) do
		if pdef.abilities and util.indexOf(pdef.abilities, "carryable") ~= nil then
			-- Check to make sure we don't add the same ability twice
			alreadyHasAbility = false
			for _, abilityName in pairs(pdef.abilities) do
				if abilityName == "itemThrow" then
					alreadyHasAbility = true
					break
				end
			end
			if not alreadyHasAbility then
				table.insert(pdef.abilities, "selectItemToss")
				if isGrenade(pdef) then
					-- Only add the deploy option to grenades
					table.insert(pdef.abilities, "selectItemDeploy")
				end
				table.insert(pdef.abilities, "itemThrow")
			end
		end
	end

	-- Hack to stop boardrig:onSimEvent() from creating an error when it tries to respond to the EV_UNIT_THROWN event
	-- Basically, the issue is that all rigs that inherit from unitrig() have a getProp() function (which boardrig:onSimEvent() expects)
	-- However, not all itemrigs inherit from unitrig, so we need to give a getProp() function to the items that don't have one.
	if itemrig.eaa_old_createRig == nil then
		itemrig.eaa_old_createRig = itemrig.createRig

		itemrig.createRig = function( boardRig, unit )
			local newrig = itemrig.eaa_old_createRig( boardRig, unit )
			if newrig.getProp == nil then
				newrig.getProp = function() return nil end
			end
			return newrig
		end

	end
end

-- LATE load/unload only. Removes throw from items that do not have throw by default
local function removeItemThrow()
	local removeIndex
	local itemdefs = include("sim/unitdefs/itemdefs")
	local propdefs = include("sim/unitdefs/propdefs")

	for name, itemdef in pairs(itemdefs) do
		if itemdef.abilities ~= nil then
			removeIndex = util.indexOf(itemdef.abilities, "selectItemToss")
			if removeIndex ~= nil then
				table.remove(itemdef.abilities, removeIndex)
			end
			removeIndex = util.indexOf(itemdef.abilities, "selectItemDeploy")
			if removeIndex ~= nil then
				table.remove(itemdef.abilities, removeIndex)
			end
			removeIndex = util.indexOf(itemdef.abilities, "itemThrow")
			if removeIndex ~= nil then
				table.remove(itemdef.abilities, removeIndex)
			end
		end
	end

	for name, propdef in pairs(propdefs) do
		if propdef.abilities and util.indexOf(propdef.abilities, "carryable") ~= nil then
			removeIndex = util.indexOf(propdef.abilities, "selectItemToss")
			if removeIndex ~= nil then
				table.remove(propdef.abilities, removeIndex)
			end
			removeIndex = util.indexOf(propdef.abilities, "selectItemDeploy")
			if removeIndex ~= nil then
				table.remove(propdef.abilities, removeIndex)
			end
			removeIndex = util.indexOf(propdef.abilities, "itemThrow")
			if removeIndex ~= nil then
				table.remove(propdef.abilities, removeIndex)
			end
		end
	end
end

-- LATE load/unload only. Restores default sightable and hidesInCover traits for items
-- Technically I should also remove eaa_doNotDestroy, but it won't hurt anything if I leave it...
local function removeItemSightable()
	for _, itemdef in ipairs(modifiedItemsAndProps.addedSightable) do
		itemdef.traits.sightable = nil
	end
	for _, itemdef in ipairs(modifiedItemsAndProps.addedHidesInCover) do
		itemdef.traits.hidesInCover = nil
	end
end

-- LATE load/unload only. Makes sightable and hidesInCover to items. This makes items visible to cameras and works with our guard brain changes to make them visible to guards as well.
-- Also add eaa_doNotDestroy to items required to complete mission (like incognita from the final mission)
local function addItemSightable(modApi, modOptions, isSetToPickupOnly)
	local itemdefs = include("sim/unitdefs/itemdefs")
	local propdefs = include("sim/unitdefs/propdefs") -- Some items are in propdefs

	removeItemSightable() -- Make sure we start with a blank slate

	for name, itemdef in pairs(itemdefs) do
		if itemdef.traits ~= nil and itemdef.traits.holoProjector ~= true and (itemdef.abilities == nil or util.indexOf(itemdef.abilities, "carryable") ~= nil) then
			if itemdef.traits.sightable == nil then
				itemdef.traits.sightable = true
				table.insert(modifiedItemsAndProps.addedSightable, itemdef)
			end
			if itemdef.traits.hidesInCover == nil then
				itemdef.traits.hidesInCover = true
				table.insert(modifiedItemsAndProps.addedHidesInCover, itemdef)
			end
		end
	end

	for name, propdef in pairs(propdefs) do
		if propdef.abilities and util.indexOf(propdef.abilities, "carryable") ~= nil and propdef.traits ~= nil then
			if propdef.traits.sightable == nil then
				propdef.traits.sightable = true
				table.insert(modifiedItemsAndProps.addedSightable, propdef)
			end
			if propdef.traits.hidesInCover == nil then
				propdef.traits.hidesInCover = true
				table.insert(modifiedItemsAndProps.addedHidesInCover, propdef)
			end
		end
	end

	-- Flag important items
	-- Vanilla items:
	if itemdefs.augment_final_level ~= nil then -- augment that gets installed in monst3r during the final level. Since this isn't droppable marking it as eaa_doNotDestroy isn't really necessary.
		itemdefs.augment_final_level.traits.eaa_doNotDestroy = true
	end
	if itemdefs.quest_material ~= nil then -- no idea what this is
		itemdefs.quest_material.traits.eaa_doNotDestroy = true
	end
	if itemdefs.special_exit_passcard ~= nil then -- passcard required to reach the teleporter
		itemdefs.special_exit_passcard.traits.eaa_doNotDestroy = true
	end
	if itemdefs.item_incognita ~= nil then -- incognita as an item. Carried by Central in the final mission.
		itemdefs.item_incognita.traits.eaa_doNotDestroy = true
	end
	-- Contingency Plan DLC items:
	if propdefs.item_incognita_case ~= nil then -- special container that must be collected during the mid mission.
		propdefs.item_incognita_case.traits.eaa_doNotDestroy = true
	end
	if propdefs.item_power_cell ~= nil then -- power cell that must be retrieved each mission (after the mid mission) to keep incognita alive.
		propdefs.item_power_cell.traits.eaa_doNotDestroy = true
	end
end

-- LATE load/unload only. Adds eaa_doNotDestroy to items that must be preserved to escape a mission
local function addDoNotDestroy()
	-- Vanilla items:
	local itemdefs = include("sim/unitdefs/itemdefs")
	if itemdefs.augment_final_level ~= nil then -- augment that gets installed in monst3r during the final level. Since this isn't droppable marking it as eaa_doNotDestroy isn't really necessary.
		itemdefs.augment_final_level.traits.eaa_doNotDestroy = true
	end
	if itemdefs.quest_material ~= nil then -- no idea what this is
		itemdefs.quest_material.traits.eaa_doNotDestroy = true
	end
	if itemdefs.special_exit_passcard ~= nil then -- passcard required to reach the teleporter
		itemdefs.special_exit_passcard.traits.eaa_doNotDestroy = true
	end
	if itemdefs.item_incognita ~= nil then -- incognita as an item. Carried by Central in the final mission.
		itemdefs.item_incognita.traits.eaa_doNotDestroy = true
	end
	-- Contingency Plan DLC items:
	-- For some reason the CP DLC items are propdefs, so we have to reference propdefs for these guys
	local propdefs = include("sim/unitdefs/propdefs")
	if propdefs.item_incognita_case ~= nil then -- special container that must be collected during the mid mission.
		propdefs.item_incognita_case.traits.eaa_doNotDestroy = true
	end
	if propdefs.item_power_cell ~= nil then -- power cell that must be retrieved each mission (after the mid mission) to keep incognita alive.
		propdefs.item_power_cell.traits.eaa_doNotDestroy = true
	end
end

----------------------------------------------------------------------
-- Functions called by the modApi
--

local function earlyInit(modApi)
	-- Some mods overwrite/add items during lateLoad, we need to make sure these mods load before ours.
	modApi.requirements = {"Function Library", "New Items And Augments", "Talon Recruitment"}
end

-- init will be called once (technically, it gets called twice because Invisible Inc. 'proofchecks' mods on startup)
local function init(modApi)
	log:write("Extra Agent Abilities - v1.2.02 -- RELEASE")
	EXTRA_AGENT_ABILITIES_SCRIPT_PATH = modApi:getScriptPath()
	local dataPath = modApi:getDataPath()

	-- Mount data.
	KLEIResourceMgr.MountPackage(dataPath .."/gui.kwad", "data")
	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/agentDoorBreak/simdefs")
	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemThrow/simdefs")

	-- Extend functions
	local itemGuardBehavior = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemGuardBehavior")
	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/agentDoorBreak/skilldefs")
	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemThrow/hud")

	-- Add generation options
	modApi:addGenerationOption("eaa_doorSlam", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_SLAM , STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_SLAM_TIP,
	{
		values = {-1, 0, 1, 2},
		value = 0,
		strings = {"Off", "0 AP", "1 AP", "2 AP"},
		noUpdate = true
	})
	modApi:addGenerationOption("eaa_doorBreak", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_BREAK, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_BREAK_TIP,
	{
		values = {-1, 0, 1, 2},
		value = 1,
		strings = {"Off", "0 AP", "1 AP", "2 AP"},
		noUpdate = true
	})
	--modApi:addGenerationOption("eaa_doorBreakMPCost", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_BREAK_MP_COST, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_DOOR_BREAK_MP_COST_TIP,
	--{
	--	values = {0, 1, 2},
	--	value = 1,
	--	strings = {"0", "1", "2"},
	--	noUpdate = true
	--})
	modApi:addGenerationOption("eaa_itemThrow", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_THROW, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_THROW_TIP,
	{
		values = {-1, 0, 1, 2},
		value = 0,
		strings = {"Off", "0 AP", "1 AP", "2 AP"},
		noUpdate = true
	})
	--modApi:addGenerationOption("eaa_itemThrowMPCost", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_THROW_MP_COST, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_THROW_MP_COST_TIP,
	--{
	--	values = {0, 1, 2},
	--	value = 0,
	--	strings = {"0", "1", "2"},
	--	noUpdate = true
	--})
	modApi:addGenerationOption("eaa_grenadeThrowMPCost", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GRENADE_THROW_MP_COST, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GRENADE_THROW_MP_COST_TIP,
	{
		values = {0, 1, 2},
		value = 0,
		strings = {"0 AP", "1 AP", "2 AP"},
		noUpdate = true
	})
	modApi:addGenerationOption("eaa_items", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEMS, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEMS_TIP, {noUpdate = true})
	modApi:addGenerationOption("eaa_itemSightable", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_SIGHTABLE, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_SIGHTABLE_TIP,
	{
		values =
		{
			itemGuardBehavior.ITEM_SIGHTABLE_VALUES.OFF,
			itemGuardBehavior.ITEM_SIGHTABLE_VALUES.PARTIAL,
			itemGuardBehavior.ITEM_SIGHTABLE_VALUES.FULL
		},
		value = itemGuardBehavior.ITEM_SIGHTABLE_VALUES.FULL,
		strings =
		{
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_SIGHTABLE_VALUES.OFF,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_SIGHTABLE_VALUES.PARTIAL,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_SIGHTABLE_VALUES.FULL
		},
		noUpdate = true
	})
	modApi:addGenerationOption("eaa_guardsSeeThrow", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_SEE_THROW, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_SEE_THROW_TIP, {noUpdate = true})
	modApi:addGenerationOption("eaa_guardsItemResponse", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_ITEM_RESPONSE, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_ITEM_RESPONSE_TIP,
	{
		values =
		{
			itemGuardBehavior.GUARDS_ITEM_RESPONSE_VALUES.INVESTIGATE,
			itemGuardBehavior.GUARDS_ITEM_RESPONSE_VALUES.MIX,
			itemGuardBehavior.GUARDS_ITEM_RESPONSE_VALUES.ALERT
		},
		value = itemGuardBehavior.GUARDS_ITEM_RESPONSE_VALUES.ALERT,
		strings =
		{
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_ITEM_RESPONSE_VALUES.INVESTIGATE,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_ITEM_RESPONSE_VALUES.MIX,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_ITEM_RESPONSE_VALUES.ALERT
		},
		noUpdate = true
	})
	modApi:addGenerationOption("eaa_itemDestroy", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY_TIP,
	{
		values =
		{
			itemGuardBehavior.ITEM_DESTROY_VALUES.VANILLA,
			itemGuardBehavior.ITEM_DESTROY_VALUES.PICKUP,
			itemGuardBehavior.ITEM_DESTROY_VALUES.PICKUP_DESTROY,
			itemGuardBehavior.ITEM_DESTROY_VALUES.DESTROY
		},
		value = itemGuardBehavior.ITEM_DESTROY_VALUES.PICKUP,
		strings =
		{
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY_VALUES.VANILLA,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY_VALUES.PICKUP,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY_VALUES.PICKUP_DESTROY,
			STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_ITEM_DESTROY_VALUES.DESTROY
		},
		noUpdate = true
	})
	modApi:addGenerationOption("eaa_guardsLootCorpses", STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_LOOT_CORPSES, STRINGS.EXTRA_AGENT_ABILITIES.OPTIONS.ENABLE_GUARDS_LOOT_CORPSES_TIP, {noUpdate = true})

end

-- load will be refreshed and run each time a new campaign is started
local function load(modApi, options, params, mod_options)
	local abilitydefs = include("sim/abilitydefs")

	EXTRA_AGENT_ABILITIES_SCRIPT_PATH = modApi:getScriptPath()
	local dataPath = modApi:getDataPath()
	local throwShared = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/throwShared")
	local itemGuardBehavior = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemGuardBehavior")

	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/doorActionsModification")

	-- Door slamming
	if options.eaa_doorSlam.value ~= -1 then
		modApi:addAbilityDef("doorSlam", EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/doorSlam/doorSlam")
		simdefs.EAA_DOOR_SLAM_MP_COST = options.eaa_doorSlam.value
		local doorSlam = abilitydefs.lookupAbility("doorSlam")
		if options.eaa_doorSlam.value > 0 then
			doorSlam.usesMP = true
		else
			doorSlam.usesMP = false
		end
	else
		simdefs.EAA_DOOR_SLAM_MP_COST = -1
	end

	-- Door breaking
	local doorBreakSkillModifications = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/agentDoorBreak/skilldefs")
	if options.eaa_doorBreak.value ~= -1 then
		modApi:addAbilityDef("agentDoorBreak", EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/agentDoorBreak/agentDoorBreak")
		simdefs.EAA_DOOR_BREAK_MP_COST = options.eaa_doorBreak.value
		doorBreakSkillModifications.addDoorBreakStrings()
		local doorBreak = abilitydefs.lookupAbility("agentDoorBreak")
		if options.eaa_doorBreak.value > 0 then
			doorBreak.usesMP = true
		else
			doorBreak.usesMP = false
		end
	else
		simdefs.EAA_DOOR_BREAK_MP_COST = -1
		doorBreakSkillModifications.removeDoorBreakStrings()
	end

	-- Item throwing
	if options.eaa_itemThrow.value ~= -1 then
		modApi:addAbilityDef("itemThrow", EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/itemThrow")
		modApi:addAbilityDef("selectItemToss", EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/selectItemToss")
		modApi:addAbilityDef("selectItemDeploy", EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/selectItemDeploy")
	end
	throwShared.setModInfo(options.eaa_itemThrow.value, options.eaa_grenadeThrowMPCost.value, options.eaa_guardsSeeThrow.enabled == true, options.eaa_itemSightable.enabled == true)

	-- Add Items
	if options.eaa_items.enabled == true then
		if options.eaa_doorBreak.value ~= -1 then
			local itemdefs = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/agentDoorBreak/agentDoorBreak_itemdefs")
			for name, itemDef in pairs(itemdefs) do
				modApi:addItemDef(name, itemDef)
			end
		end
	end

	-- Guard behavior
	itemGuardBehavior.setModInfo(modApi, dataPath,
	                             options.eaa_itemSightable.value,
	                             options.eaa_guardsItemResponse.value,
	                             options.eaa_itemDestroy.value,
	                             options.eaa_guardsLootCorpses.enabled == true)
end

local function unload(modApi, options)
	local throwShared = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/itemThrow/throwShared")
	local doorBreakSkillModifications = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/agentDoorBreak/skilldefs")
	local itemGuardBehavior = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemGuardBehavior")

	throwShared.unload()
	doorBreakSkillModifications.removeDoorBreakStrings()
	itemGuardBehavior.unload()
end

local function lateLoad(modApi, options)
	local scriptPath = modApi:getScriptPath()
	local itemGuardBehavior = include(scriptPath.."/itemGuardBehavior")
	include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH.."/itemThrow/throwExtender") -- Must be in lateLoad to catch other mods throw abilities.

	-- Must come in lateLoad to catch all of the agents and items from DLCs and mods
	if options.eaa_doorSlam.value ~= -1 then
		addDoorSlam()
	else
		removeDoorSlam()
	end

	if options.eaa_doorBreak.value ~= -1 then
		addAgentDoorBreak()
	else
		removeAgentDoorBreak()
	end

	if options.eaa_itemThrow.value ~= -1 then
		addItemThrow()
	else
		removeItemThrow()
	end

	if options.eaa_itemSightable.value ~= itemGuardBehavior.ITEM_SIGHTABLE_VALUES.OFF then
		addItemSightable(modApi, options, options.eaa_itemDestroy.value == itemGuardBehavior.ITEM_DESTROY_VALUES.PICKUP)
	else
		removeItemSightable()
	end

	if options.eaa_itemDestroy ~= itemGuardBehavior.ITEM_DESTROY_VALUES.VANILLA then
		addDoNotDestroy()
	end
end

-- lateUnload is needed to be able to remove doorSlam/itemThrow/etc. from DLC/mod agents/items
local function lateUnload(modApi, options)
	removeDoorSlam()
	removeAgentDoorBreak()
	removeItemThrow()
	removeItemSightable()
end


-- Called once, tells the program where to look for strings at
local function initStrings(modApi)
	EXTRA_AGENT_ABILITIES_SCRIPT_PATH = modApi:getScriptPath()
	local dataPath = modApi:getDataPath()

	local MOD_STRINGS = include(EXTRA_AGENT_ABILITIES_SCRIPT_PATH .."/strings")
	modApi:addStrings(dataPath, "EXTRA_AGENT_ABILITIES", MOD_STRINGS) -- This allows us to refer to our strings by STRINGS.EXTRA_AGENT_ABILITIES....
end

return {
	earlyInit = earlyInit,
	init = init,
	load = load,
	unload = unload,
	lateLoad = lateLoad,
	lateUnload = lateUnload,
	initStrings = initStrings
}
